// =======================================================================
// 
//       Filename:  commandelement.cpp
// 
//    Description:  
// 
//        Version:  1.0
//        Created:  20.04.2010 00:55:51
//       Revision:  none
//       Compiler:  g++
// 
//         Author:  Dominik 'domachine' Burgdörfer (-), dominik.burgdoerfer@googlemail.com
//        Company:  -
// 
// =======================================================================

#include "command.hpp"
#include "../symbols/function.hpp"
#include "../symbols/macro.hpp"
#include "objectwrapper.hpp"
#include "../callinfo.hpp"
#include "../elementstack.hpp"
#include "../exceptions/processingerror.hpp"
#include "../types/object.hpp"
#include "objectwrapper.hpp"
#include <stdexcept>
#include <sstream>

using namespace std;
using namespace boost;

namespace wesl {
    namespace elements {
        Command::Command(ElementStack& parent)
            : Element(parent),
            Callable(this),
            Resolvable(this),
            Printable(this)
        {

        }

        Command::Command(ElementStack& parent, const string& name,
                    vector<shared_ptr<elements::Element> >& params)
            : Element(parent),
            Callable(this),
            Resolvable(this),
            Printable(this),
            mName(name), mParams(params)
        {

        }

        string Command::name() const
        {
            return mName;
        }

        void Command::setName(const string& name)
        {
            mName = name;
        }

        vector<shared_ptr<elements::Element> >& Command::parameters()
        {
            return mParams;
        }

        void Command::resolveParameters(
                vector<elements::ElementPtr>& in,
                vector<types::ObjectPtr>& out)
        {
            string cmdName = name();
            unsigned int i;

            for(i = 0; i < in.size(); ++i)
            {
                shared_ptr<Element> element = in[i];

                if(element->is(Element::RESOLVABLE)) {
                    ResolvablePtr resolvablePtr =
                        dynamic_pointer_cast<Resolvable>(element);
                    out.push_back(resolvablePtr->resolve());
                }
                else {
                    ostringstream msg;
                    msg << "argument " << i
                        << " is not resolvable to an object";
                    throw exceptions::ProcessingError(msg.str(),
                            parent().sourcePath(),
                            line());
                }

                if(out.back()->objectType() == types::Object::VOID) {
                    ostringstream msg;
                    
                    msg << cmdName << ": invalid type " 
                            << "Void for argument " 
                            << out.size();
                    throw exceptions::ProcessingError(msg.str(),
                            parent().sourcePath(),
                            line());
                }
            }
        }

        void Command::print(ostream& stream) {
            string cmdName = name();

            CallInfo::ObjectPointerList args;

            resolveParameters(parameters(), args);

            CallInfo callInfo(args, *this);

            ElementStack::MacroMap::iterator
                funcIter = parent().lookupMacro(cmdName);

            if(funcIter != ElementStack::STATIC_MACRO_MAP.end()) {
                funcIter->second->call(callInfo, stream);
            }
            else {
                resolve()->print(stream);
            }
        }

        ObjectPtr Command::resolve() {
            string cmdName = name();
            types::ObjectPtr elementPtr;

            CallInfo::ObjectPointerList args;

            resolveParameters(parameters(), args);

            CallInfo callInfo(args, *this);

            ElementStack::FunctionMap::iterator
                funcIter = parent().lookupFunction(cmdName);

            if(funcIter != ElementStack::STATIC_FUNCTION_MAP.end()) {
                elementPtr = 
                        funcIter->second->call(callInfo);
            }
            else
                throw exceptions::ProcessingError(
                        "not a function: " + 
                        cmdName, parent().sourcePath(),
                        line());

            return elementPtr;
        }

        void Command::call() {
            string cmdName = name();

            CallInfo::ObjectPointerList args;

            resolveParameters(parameters(), args);

            CallInfo callInfo(args, *this);

            ElementStack::FunctionMap::iterator
                funcIter = parent().lookupFunction(cmdName);

            if(funcIter != ElementStack::STATIC_FUNCTION_MAP.end())
                funcIter->second->call(callInfo);
            else
                throw exceptions::ProcessingError(
                        "not a function: " + 
                        cmdName, parent().sourcePath(),
                        line());

        }
    }
}
