// =======================================================================
// 
//       Filename:  elementstack.cpp
// 
//    Description:  
// 
//        Version:  1.0
//        Created:  06.06.2010 18:35:32
//       Revision:  none
//       Compiler:  g++
// 
//         Author:  Dominik 'domachine' Burgdörfer (-), dominik.burgdoerfer@googlemail.com
//        Company:  -
// 
// =======================================================================

#include "elementstack.hpp"
#include "types/string.hpp"
#include "exceptions/processingerror.hpp"
#include "exceptions/returnexception.hpp"
#include "types/void.hpp"
#include "types/integer.hpp"
#include "types/float.hpp"
#include "types/string.hpp"
#include "elements/objectwrapper.hpp"
#include "elements/callable.hpp"
#include <boost/foreach.hpp>
#include <sstream>
#define foreach BOOST_FOREACH

using namespace std;
using namespace boost;
using namespace wesl::elements;
using wesl::exceptions::ProcessingError;

namespace wesl {
    ElementStack::ElementStack()
        : Element(), 
        Printable(this),
        Resolvable(this),
        m_processed(false)
    {

    }

    ElementStack::ElementStack(const boost::filesystem::path& sourcePath)
        : Element(), 
        Printable(this),
        Resolvable(this),
        m_sourcePath(sourcePath), 
        m_processed(false)
    {

    }

    void ElementStack::open() {
        m_input.open(m_sourcePath.string().c_str());

        m_iterator.setStream(m_input);
    }

    void ElementStack::close() {
        if(m_input.is_open())
            m_input.close();
    }

    utils::CIterator& ElementStack::iterator() {
        return m_iterator;
    }

    boost::filesystem::path ElementStack::sourcePath() const
    {
        return m_sourcePath;
    }

    void ElementStack::setProcessed(bool state) {
        m_processed = true;
    }

    bool ElementStack::processed() const
    {
        return m_processed;
    }

    ElementStack& ElementStack::head() {
        if(hasParent())
            return parent().head();
        else
            return *this;
    }

    ElementStack::VariableMap::iterator ElementStack::lookupVariable(
            const string& name)
    {
        shared_ptr<types::Object> object;

        ElementStack::VariableMap::iterator iter;
        ElementStack::VariableMap::iterator endIter;
        ElementStack* parent;
        bool firstRun = true;

        do {
            if(firstRun) {
                parent = this;
                firstRun = false;
            }
            else
                parent = &this->parent();

            iter =
                parent->dynamicVariableMap().find(name);
            endIter =
                parent->dynamicVariableMap().end();

            if(iter != endIter)
                return iter;

        } while(parent->hasParent());

        iter = 
            ElementStack::STATIC_VARIABLE_MAP.find(name);
        endIter = 
            ElementStack::STATIC_VARIABLE_MAP.end();

        return iter;
    }

    ElementStack::VariableMap& ElementStack::dynamicVariableMap() {
        return m_dynamicVariableMap;
    }

    const ElementStack::VariableMap& ElementStack::dynamicVariableMap() const
    {
        return m_dynamicVariableMap;
    }

    ElementStack::FunctionMap::iterator ElementStack::lookupFunction(
            const string& name)
    {
        ElementStack::FunctionMap& dynamicMap =
            head().dynamicFunctionMap();

        ElementStack::FunctionMap::iterator
            funcIter = 
            dynamicMap.find(name);

        if(funcIter != dynamicMap.end())
            return funcIter;
        else {
            funcIter = 
                ElementStack::STATIC_FUNCTION_MAP.find(name);

            if(funcIter != ElementStack::STATIC_FUNCTION_MAP.end())
                return funcIter;
            else {
                funcIter = 
                    ElementStack::STATIC_FUNCTION_MAP.find("*");

                return funcIter;
            }
        }
    }

    ElementStack::FunctionMap& ElementStack::dynamicFunctionMap() {
        return m_dynamicFunctionMap;
    }

    const ElementStack::FunctionMap& ElementStack::dynamicFunctionMap() const
    {
        return m_dynamicFunctionMap;
    }

    ElementStack::MacroMap::iterator ElementStack::lookupMacro(
            const string& name)
    {
        ElementStack::MacroMap& dynamicMap =
            head().dynamicMacroMap();

        ElementStack::MacroMap::iterator
            macroIter = 
            dynamicMap.find(name);

        if(macroIter != dynamicMap.end())
            return macroIter;
        else {
            macroIter = 
                ElementStack::STATIC_MACRO_MAP.find(name);

            if(macroIter != ElementStack::STATIC_MACRO_MAP.end())
                return macroIter;
            else {
                macroIter = 
                    ElementStack::STATIC_MACRO_MAP.find("*");

                return macroIter;
            }
        }
    }

    ElementStack::MacroMap& ElementStack::dynamicMacroMap() {
        return m_dynamicMacroMap;
    }

    const ElementStack::MacroMap& ElementStack::dynamicMacroMap() const
    {
        return m_dynamicMacroMap;
    }

    void ElementStack::print(ostream& stream) {
        open();
        

        foreach(shared_ptr<Element> current, 
                *this)
        {
        
            if(current->is(Element::PRINTABLE)) {
        
                PrintablePtr ptr = 
                    dynamic_pointer_cast<Printable>(current);
        
                ptr->print(stream);
            }
            else if(current->is(Element::RESOLVABLE)) {
        
                ResolvablePtr ptr =
                    dynamic_pointer_cast<Resolvable>(current);
        
                ptr->resolve()->print(stream);
        
            }
            else if(current->is(Element::CALLABLE)) {
        
                CallablePtr ptr =
                    dynamic_pointer_cast<Callable>(current);
        
                ptr->call();
        
            }
            else
                throw ProcessingError("Internal processing-error",
                              sourcePath(),
                              current->line());
        }

        if(!hasParent()) {
            dynamicVariableMap().clear();
            dynamicFunctionMap().clear();
            dynamicMacroMap().clear();
        }
        
        setProcessed();
        close();
    }

    types::ObjectPtr ElementStack::resolve() {
        types::VoidPtr voidPtr(new types::Void());
        types::ObjectPtr elementPtr(voidPtr);
        open();

        foreach(shared_ptr<Element> current, 
                *this)
        {
            try {
                if(current->is(Element::CALLABLE)) {
                    CallablePtr ptr =
                        dynamic_pointer_cast<Callable>(current);
                    ptr->call();
                }
                else
                    throw ProcessingError("Element is not callable",
                            sourcePath(),
                            current->line());
            }
            catch(exceptions::ReturnException& e) {
                elementPtr = e.object();
            }
        }

        dynamicVariableMap().clear();
        dynamicFunctionMap().clear();
        dynamicMacroMap().clear();
        
        setProcessed();
        close();

        return elementPtr;
    }

    ElementStack::FunctionMap ElementStack::STATIC_FUNCTION_MAP =
        ElementStack::FunctionMap();

    ElementStack::MacroMap ElementStack::STATIC_MACRO_MAP =
        ElementStack::MacroMap();

    ElementStack::VariableMap ElementStack::STATIC_VARIABLE_MAP =
        ElementStack::VariableMap();
}
