// =======================================================================
// 
//       Filename:  parser.cpp
// 
//    Description:  
// 
//        Version:  1.0
//        Created:  26.03.2010 19:52:09
//       Revision:  none
//       Compiler:  g++
// 
//         Author:  Dominik 'domachine' Burgdörfer (-), dominik.burgdoerfer@googlemail.com
//        Company:  -
// 
// =======================================================================

#include "parser.hpp"
#include "utils/citerator.hpp"
#include "utils/regexitermatch.hpp"
#include "exceptions/syntaxerror.hpp"
#include "elements/command.hpp"
#include "elements/text.hpp"
#include "elements/script.hpp"
#include "elements/variablerecall.hpp"
#include "symbols/function.hpp"
#include "symbols/macro.hpp"
#include "elements/returnstatement.hpp"
#include "elements/subroutinedefinition.hpp"
#include "elements/objectwrapper.hpp"
#include "types/float.hpp"
#include "types/integer.hpp"
#include "types/void.hpp"
#include <boost/regex.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <iostream>
#include <fstream>

using namespace std;
using namespace boost;
using namespace wesl::elements;
using namespace wesl::symbols;
using wesl::exceptions::SyntaxError;
namespace fs = boost::filesystem;

namespace wesl {
    Parser::Parser()
        : m_line(0)
    {

    }

    void Parser::parse(ElementStack& elementStack, int begin)
    {
        utils::CIterator bIterator(mIterator);
        ifstream input;

        input.open(elementStack.sourcePath().string().c_str());
        mIterator.setStream(input);

        m_line = 1;
        parseElementStack(elementStack, begin);

        input.close();
        mIterator = bIterator;
    }

    void Parser::parseElementStack(ElementStack& elementStack, 
                                   int begin, bool ignoreText)
    {
        while(mIterator.position() < begin) {
            safelyIncr();
            if(mIterator.currentChar() == '\n')
                ++m_line;
        }

        m_filePath = elementStack.sourcePath();

        bool textBlock = false;
        Text* textElement = 0;

        for(; mIterator.good(); mIterator.moveForward()) {
            if(mIterator.currentChar() != '{') {
                if(!textBlock) {
                    textElement = new Text(
                            elementStack,
                            mIterator.position(), -1);
                    textElement->setLine(m_line);

                    textBlock = true;
                }

                if(mIterator.currentChar() == '\n')
                    ++m_line;
            }
            else {
                if(mIterator.nextChar() == '{') {
                    if(textBlock && !ignoreText) {
                        textElement->setEnd(mIterator.position());
                        safelyIncr();
                        elementStack.push_back(
                                shared_ptr<Text>(
                                    textElement));
                        textElement = new Text(
                                elementStack,
                                mIterator.position(), -1);
                        textElement->setLine(m_line);
                        continue;
                    }
                }

                if(textBlock && !ignoreText) {
                    textElement->setEnd(mIterator.position());
                    elementStack.push_back(
                            shared_ptr<Text>(
                                textElement));
                    textBlock = false;
                }

                shared_ptr<Element> element = parseDynamicContent(
                                                          elementStack);
                bool isDnlCmd = false;
                if(typeid(*element) == typeid(Command)) {
                    shared_ptr<Command> cmd =
                        dynamic_pointer_cast<Command>(element);
                    if(cmd->name() == "dnl") {
                        safelyIncr();
                        if(mIterator.currentChar() != '\n')
                            throw SyntaxError(
                                    "dnl: no newline to delete",
                                    cmd->line(),
                                    m_filePath);

                        ++m_line;
                        isDnlCmd = true;
                    }
                    else if(cmd->name() == "end") {
                        break;
                    }
                }
                if(!isDnlCmd && !ignoreText)
                    elementStack.push_back(element);
            }
        }

        if(textElement != 0 && textElement->end() == -1 && !ignoreText) 
        {
            textElement->setEnd(mIterator.position());
            elementStack.push_back(
                    shared_ptr<Text>(
                        textElement));
        }
    }

    int Parser::line() const
    {
        return m_line;
    }

    void Parser::safelyIncr()
    {
        mIterator.moveForward();

        if(!mIterator.good())
            throw SyntaxError("unexpected end-of-stream.",
                    line(), m_filePath);
    }

    void Parser::skipWhiteSpace()
    {
        regex spaceregex("[ \t\n]");

        while(utils::regexIterMatch(mIterator, spaceregex)) {
            if(mIterator.currentChar() == '\n')
                ++m_line;

            safelyIncr();
        }
    }

    ElementPtr
                Parser::parseDynamicContent(ElementStack& elementStack)
    {
        safelyIncr();

        skipWhiteSpace();

        if(mIterator.currentChar() == '!') {
            return parseScript(elementStack);
        }
        else if(mIterator.currentChar() == '$') {
            return parseVar(elementStack);
        }
        else if(mIterator.currentChar() == ':') {
            shared_ptr<Text> text(new Text(elementStack));
            text->setLine(line());

            safelyIncr();
            skipWhiteSpace();

            string pathRaw = static_cast<string>(
                    *parseString());
            fs::path path(pathRaw);

            safelyIncr();

            skipWhiteSpace();
            if(mIterator.currentChar() != '}')
                throw SyntaxError("unexpected char, expected '}'", line(),
                        m_filePath);

            text->setExternalFile(path);

            return text;
        }
        else if(mIterator.currentChar() == '#') {
            int lineNo = line();

            safelyIncr();
            skipWhiteSpace();

            string pathRaw = static_cast<string>(
                    *parseString());
            fs::path path(pathRaw);

            shared_ptr<ElementStack> stack(new ElementStack(path));
            stack->setLine(lineNo);
            stack->setParent(elementStack);

            safelyIncr();

            skipWhiteSpace();
            if(mIterator.currentChar() != '}')
                throw SyntaxError("unexpected char, expected '}'", line(),
                        m_filePath);

            Parser subParser;
            subParser.parse(*stack);
            return stack;
        }
        else {
            string cmdName;
            regex cmdregex("[a-zA-Z_0-9]");
            vector<string> args;

            while(utils::regexIterMatch(mIterator, cmdregex)) {
                cmdName.push_back(mIterator.currentChar());
                safelyIncr();
            }

            if(cmdName.size() == 0)
                throw SyntaxError("no or no valid command name",
                        line(), m_filePath);

            if(cmdName == "function" || cmdName == "macro") {
                string keyword = cmdName;
                skipWhiteSpace();

                cmdName.clear();
                while(utils::regexIterMatch(mIterator, cmdregex)) {
                    cmdName.push_back(mIterator.currentChar());
                    safelyIncr();
                }

                if(cmdName.size() == 0)
                    throw SyntaxError("no or no valid function name",
                            line(), m_filePath);

                skipWhiteSpace();

                if(mIterator.currentChar() == ':') {
                    safelyIncr();

                    skipWhiteSpace();

                    string currentArg;
                    while(true) {
                        if(mIterator.currentChar() == '}')
                            break;
                        else if(mIterator.currentChar() == ',') {
                            safelyIncr();
                            skipWhiteSpace();
                        }
                        else {
                            currentArg.clear();
                            while(utils::regexIterMatch(mIterator, 
                                                            cmdregex))
                            {
                                currentArg.push_back(
                                        mIterator.currentChar());
                                safelyIncr();
                            }
                            args.push_back(currentArg);
                            skipWhiteSpace();
                        }
                    }
                }
                else if(mIterator.currentChar() != '}') {
                    throw SyntaxError(
                            "unexpected char, expected '}' or ':'",
                            line(), m_filePath);
                }

                safelyIncr();

                SubroutinePtr subroutinePtr;
                    
                if(keyword == "function") {
                    subroutinePtr = SubroutinePtr(new Function(
                                            elementStack.sourcePath(),
                                            args));
                    parseElementStack(subroutinePtr->body(), 
                                  mIterator.position(),
                                  true);
                }
                else if(keyword == "macro") {
                    subroutinePtr = SubroutinePtr(new Macro(
                                            elementStack.sourcePath(),
                                            args));
                    parseElementStack(subroutinePtr->body(), 
                                  mIterator.position());
                }
                else
                    throw logic_error("unknow error: keyword change "
                            "during parsing process");

                subroutinePtr->body().setParent(elementStack);

                
                subroutinePtr->body().parent();

                SubroutineDefinitionPtr subroutineDefPtr(
                        new SubroutineDefinition(cmdName, subroutinePtr));

                return subroutineDefPtr;
            }
            else if(cmdName == "return") {
                shared_ptr<ReturnStatement> returnStmt(
                        new ReturnStatement(elementStack));
                returnStmt->setLine(line());
                skipWhiteSpace();

                ElementPtr elementPtr;

                if(mIterator.currentChar() == '{')
                    elementPtr = parseDynamicContent(elementStack);
                else if(mIterator.currentChar() == '"' ||
                        mIterator.currentChar() == '\'')
                    elementPtr = wrapObject(parseString());
                else
                    elementPtr = wrapObject(parseNumber());
                
                safelyIncr();
                skipWhiteSpace();

                returnStmt->setElement(elementPtr);
                if(mIterator.currentChar() != '}')
                    throw SyntaxError("unexpected char, expected '}'", line(),
                            m_filePath);

                return returnStmt;
            }
            else
                return parseCommand(elementStack, cmdName);
        }
    }

    shared_ptr<Script> Parser::parseScript(
            ElementStack& elementStack)
    {
        regex scriptregex("[a-z0-9\\-_A-Z]");
        string scriptSpec;
        shared_ptr<Script> scriptElement( 
                new Script(elementStack));
        bool sourceScript = false;

        scriptElement->setLine(line());

        safelyIncr();
        skipWhiteSpace();
        if(mIterator.currentChar() == ':') {
            sourceScript = true;
            safelyIncr();
        }
        skipWhiteSpace();

        while(utils::regexIterMatch(mIterator, scriptregex)) {
            scriptSpec.push_back(mIterator.currentChar());
            safelyIncr();
        }

        if(scriptSpec.size() == 0)
            throw SyntaxError("no or invalid script specification",
                    line(), m_filePath);

        scriptElement->setSpecification(scriptSpec);

        skipWhiteSpace();
        if(sourceScript) {
            string pathRaw = static_cast<string>(
                    *parseString());
            fs::path path(pathRaw);
            scriptElement->setExternalFile(path);
        }

        if(!sourceScript)
            scriptElement->setBegin(mIterator.position());

        while(true) {
            if(mIterator.currentChar() == '!' && 
                    mIterator.nextChar() == '}') {

                if(!sourceScript)
                    scriptElement->setEnd(mIterator.position());

                safelyIncr();

                break;
            }
            else if(!sourceScript) {
                char cChar = mIterator.currentChar();
                if(cChar == '\n')
                    m_line++;
            }
            safelyIncr();
        }

        return scriptElement;
    }

    shared_ptr<VariableRecall> Parser::parseVar(ElementStack& elementStack)
    {
        string cmdName;
        regex cmdregex("[a-zA-Z_0-9]");
        shared_ptr<VariableRecall> varElement(
                    new VariableRecall(elementStack));

        varElement->setLine(line());

        safelyIncr();
        skipWhiteSpace();
        while(utils::regexIterMatch(mIterator, cmdregex)) {
            cmdName.push_back(mIterator.currentChar());
            safelyIncr();
        }

        if(cmdName.size() == 0)
            throw SyntaxError("no or no valid variable name",
                    line(), m_filePath);

        varElement->setName(cmdName);

        skipWhiteSpace();

        if(mIterator.currentChar() != '}')
            throw SyntaxError(
                    "unexpected char, expected '}'",
                    line(), m_filePath);
        return varElement;
    }

    shared_ptr<Command> 
        Parser::parseCommand(ElementStack& elementStack,
                             const string& cmdName) 
    {
        shared_ptr<Command> element(new Command(elementStack));

        element->setLine(line());
        element->setName(cmdName);

        skipWhiteSpace();
        
        vector<shared_ptr<Element> >& args =
                                        element->parameters();

        if(mIterator.currentChar() == ':') {
            safelyIncr();

            skipWhiteSpace();

            string currentArg;
            while(true) {
                if(mIterator.currentChar() == '"' ||
                        mIterator.currentChar() == '\'') {
                    args.push_back(wrapObject(parseString()));
                    safelyIncr();
                    skipWhiteSpace();
                }
                else if(mIterator.currentChar() == '{') {
                    args.push_back(parseDynamicContent(
                                elementStack));

                    safelyIncr();
                    skipWhiteSpace();
                }
                else if(mIterator.currentChar() == '}')
                    break;
                else if(mIterator.currentChar() == ',') {
                    safelyIncr();
                    skipWhiteSpace();
                }
                else {
                    args.push_back(wrapObject(parseNumber()));
                    skipWhiteSpace();
                }
            }
        }
        else if(mIterator.currentChar() != '}') {
            throw SyntaxError(
                    "unexpected char, expected '}' or ':'",
                    line(), m_filePath);
        }

        return element;
    }

    shared_ptr<types::String> Parser::parseString() {
        char delimiter;
        shared_ptr<types::String>
            buffer(new types::String(""));

        if(mIterator.currentChar() == '\'')
            delimiter = '\'';
        else if(mIterator.currentChar() == '"')
            delimiter = '"';
        else
            throw SyntaxError("unexpected char, expected ' or \"", line(),
                                m_filePath);

        char charBeforeLast;

        do {
            char currentChar;
            charBeforeLast = mIterator.lastChar();
            safelyIncr();

            if(mIterator.currentChar() == '\\')
                if(mIterator.lastChar() != '\\')
                    continue;

            if(mIterator.currentChar() == delimiter) {
                if(mIterator.lastChar() != '\\' || charBeforeLast == '\\')
                    break;
            }

            if(mIterator.lastChar() == '\\' && charBeforeLast != '\\') {
                if(mIterator.currentChar() == 'n')
                    currentChar = '\n';
                else if(mIterator.currentChar() == 't')
                    currentChar = '\t';
                else {
                    string errMsg("invalid escape sequence: \\");
                    errMsg.push_back(mIterator.currentChar());
                    throw SyntaxError(errMsg,
                            line(), m_filePath);
                }
            }
            else
                currentChar = mIterator.currentChar();

            buffer->push_back(currentChar);

            if(mIterator.currentChar() == '\n') {
                if(mIterator.lastChar() == '\\' &&
                        charBeforeLast != '\\')
                    buffer->erase(buffer->end()-1);
                ++m_line;
            }
        } while(true);

        return buffer;
    }

    shared_ptr<types::Object> Parser::parseNumber() {
        regex numRegex("[0-9\\.\\-]");
        string buffer;

        while(utils::regexIterMatch(mIterator, numRegex)) {
            buffer.push_back(mIterator.currentChar());
            safelyIncr();
        }

        if(buffer.empty())
            throw SyntaxError("unexpected token", 
                    line(), m_filePath);
        else {
            regex floatRegex("\\-?[0-9]+\\.[0-9]+");
            regex intRegex("\\-?[0-9]+");

            if(regex_match(buffer, floatRegex)) {
                double f = atof(buffer.c_str());

                if(f == HUGE_VAL || f == -HUGE_VAL)
                    throw SyntaxError("double value out of range",
                            line(), m_filePath);

                shared_ptr<types::Float> floatObj(
                        new types::Float(f));

                return floatObj;
            }
            else if(regex_match(buffer, intRegex)) {
                int i = atoi(buffer.c_str());

                if(i == INT_MAX || i == INT_MIN)
                    throw SyntaxError("integer value out of range",
                            line(), m_filePath);

                shared_ptr<types::Integer> 
                    integer(new types::Integer(i));
                return integer;
            }
            else 
                throw SyntaxError("unexpected token", 
                        line(), m_filePath);
        }
    }
}
