// =======================================================================
// 
//       Filename:  plodder.cpp
// 
//    Description:  
// 
//        Version:  1.0
//        Created:  27.03.2010 23:16:24
//       Revision:  none
//       Compiler:  g++
// 
//         Author:  Dominik 'domachine' Burgdörfer (-), dominik.burgdoerfer@googlemail.com
//        Company:  -
// 
// =======================================================================

#include "plodder.hpp"
#include "processor.hpp"
#include "pluginmanager.hpp"
#include "page/page.hpp"
#include "page/parser.hpp"
#include "utils/comparenodes.hpp"
#include "exceptions/systemerror.hpp"
#include "navigation/node.hpp"
#include "utils/pathlist.hpp"
#include "configparser.hpp"
#include <wesl/elements/script.hpp>
#include <wesl/elementstack.hpp>
#include <wesl/exceptions/syntaxerror.hpp>
#include <wesl/exceptions/fileerror.hpp>
#include <wesl/exceptions/processingerror.hpp>
#include <wesl/parser.hpp>
#include <string.h>
#include <vector>
#include <boost/regex.hpp>
#include <unistd.h>
#include <fstream>
#include <cerrno>
#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>

#define foreach BOOST_FOREACH

namespace fs = boost::filesystem;
using namespace std;
using namespace boost;
using namespace wesl::exceptions;
using namespace webplodder::exceptions;
using wesl::ElementStack;

namespace webplodder {
    Plodder::Plodder(const fs::path& configDir)
        : m_configDir(configDir)
    {

    }

    Plodder::~Plodder() {
    }

    void Plodder::init()
    {
        fs::path configFile = m_configDir / "config.yaml";

        // Open config file and parse it
        ConfigParser configParser;
        m_processor = shared_ptr<Processor>(
                new Processor());
        
        try {
            try {
                configParser.parseScriptRules(configFile,
                           wesl::elements::Script::RULES);
            }
            catch(FileError& e) {
                throw SystemError("configparser: " + 
                        e.whatStr() + ": " + 
                        formatPath(e.file()));
            }
        }
        catch(SystemError& e) {
            cerr << e.what() << endl;
        }

        // Pass the parsed config to the Processor
        
        shared_ptr<PluginManager> pluginManagerPtr(
                new PluginManager(m_configDir / "plugins"));
        
        try {
            pluginManagerPtr->scanPlugins();
        }
        catch(Exception& e) {
            cerr << "pluginmanager: " << e.what() << endl;
        }

        m_processor->installPluginManager(
                                pluginManagerPtr);
    }

    void Plodder::scanDir(const fs::path& dir,
                          utils::PathList& exclude) 
    {
        m_targetPath = dir;
        if(chdir(dir.string().c_str()) < 0)
            throw SystemError("chdir failed for: " + dir.string());

        // Open the given directory and search it for
        // page files
        fs::recursive_directory_iterator dIter(".");
        fs::recursive_directory_iterator endIter;
        vector<shared_ptr<page::Page> > pageFiles;
        fs::path templateFile;
        vector<navigation::Node> navi;
        page::Parser pageParser;

        for(; dIter != endIter; ++dIter) {
            if(fs::is_regular_file(dIter->status()) &&
                    dIter->path().extension() == ".page" &&
                    (!exclude.contains(dIter->path()))) {
                fs::path outFile(dIter->path());
                outFile.replace_extension("html");
                shared_ptr<page::Page> pageFile = 
                      shared_ptr<page::Page>(new page::Page(
                                                    dIter->path(), 
                                                    outFile));

                try {
                    // Parse the current page file ...
                    pageParser.parse(*pageFile);
                }
                catch(FileError& e) {
                    throw SystemError(e.whatStr() + ": " + 
                                           formatPath(e.file()));
                }
                catch(YAML::Exception& e) {
                    throw ParsingError(formatPath(pageFile->sourcePath()),
                                       e.what());
                }

                if(pageFile->inMenu()) {
                    navi.push_back(navigation::Node(pageFile->title(),
                                        pageFile->outputPath(),
                                        pageFile->pos()));
                }

                pageFiles.push_back(pageFile);
            }
            else if(fs::is_regular_file(dIter->status()) &&
                    dIter->path().filename() == "default.template" &&
                    dIter->path().parent_path().filename() == "." &&
                    (!exclude.contains(dIter->path())))
            {
                templateFile = dIter->path();
            }
            else if(fs::is_directory(dIter->status()) &&
                    exclude.contains(dIter->path())) {
                dIter.no_push();
            }
        }

        if(templateFile.empty())
            throw Exception("no template file found");

        if(pageFiles.empty())
            throw Exception("no page files found");

        // Sort navigation
        sort(navi.begin(),
                navi.end(), utils::compareNodes);

        // Install all navigations
        foreach(shared_ptr<page::Page> file, pageFiles)
            file->installNavigation(navi);

        // Construct a templatefile::File object
        // upon the source file and parse it
        wesl::ElementStack cTemplateFile(templateFile);
        wesl::Parser parser;

        try {
            parser.parse(cTemplateFile);
        }
        catch(SyntaxError& e) {
            throw SyntaxError(e.what(), e.line(),
                    formatPath(e.file()));
        }
        catch(FileError& e) {
            throw SystemError(e.whatStr() + ": " + 
                                    formatPath(e.file()),
                                  e.errNo());
        }

        // Iterate through the page queue
        foreach(shared_ptr<page::Page> pageFile, pageFiles)
        {
            // ... and pass it to the Processor
            try {
                pageParser.parseBody(*pageFile, 
                                     cTemplateFile.dynamicVariableMap());

                m_processor->process(*pageFile, cTemplateFile);
                pageFile->content()->clear();
            }
            catch(SyntaxError& e) {
                throw SyntaxError(e.what(), e.line(),
                        formatPath(e.file()));
            }
            catch(ProcessingError& e) {
                fs::remove(pageFile->outputPath());
                throw ProcessingError(e.what(), 
                        formatPath(e.file()),
                        e.line());
            }
            catch(FileError& e) {
                throw SystemError(e.whatStr() + ": " + 
                                    formatPath(e.file()),
                                  e.errNo());
            }
            catch(YAML::Exception& e) {
                throw ParsingError(formatPath(pageFile->sourcePath()),
                        e.what());
            }
        }
    }

    string Plodder::formatPath(const fs::path& filePath) {
        fs::path printablePath(m_targetPath);
        fs::path::iterator filePathIter = filePath.begin();

        if(filePathIter == filePath.end())
            return "";

        if(*filePathIter == "/")
            return filePath.string();
        else if(*filePathIter == ".")
            ++filePathIter;
        
        for(; filePathIter != filePath.end(); ++filePathIter)
            printablePath /= *filePathIter;

        return printablePath.string();
    }
}
